@@ -1,7 +1,9 @@
 #
-# Copyright (c) Spyder Project Contributors)
-# Licensed under the terms of the MIT License)
-# (see spyder/__init__.py for details)
+# Copyright (c) 2009- Spyder Kernels Contributors
+#
+# Licensed under the terms of the MIT License
+# (see spyder_kernels/__init__.py for details)
+# -----------------------------------------------------------------------------
 #
 # IMPORTANT NOTE: Don't add a coding line here! It's not necessary for
 # site files
@@ -15,11 +17,17 @@ import io
 import os
 import os.path as osp
 import pdb
+import re
 import shlex
 import sys
+import sysconfig
 import time
 import warnings
 
+from IPython.core.getipython import get_ipython
+
+
+# We are in Python 2?
 PY2 = sys.version[0] == '2'
 
 
@@ -121,36 +129,6 @@ if os.name == 'nt' and PY2:
         pass
 
 
-#==============================================================================
-# Cython support
-#==============================================================================
-RUN_CYTHON = os.environ.get("SPY_RUN_CYTHON") == "True"
-HAS_CYTHON = False
-
-if RUN_CYTHON:
-    try:
-        __import__('Cython')
-        HAS_CYTHON = True
-    except Exception:
-        pass
-
-    if HAS_CYTHON:
-        # Import pyximport to enable Cython files support for
-        # import statement
-        import pyximport
-        pyx_setup_args = {}
-
-        # Add Numpy include dir to pyximport/distutils
-        try:
-            import numpy
-            pyx_setup_args['include_dirs'] = numpy.get_include()
-        except Exception:
-            pass
-
-        # Setup pyximport and enable Cython files reload
-        pyximport.install(setup_args=pyx_setup_args, reload_support=True)
-
-
 #==============================================================================
 # Prevent subprocess.Popen calls to create visible console windows on Windows.
 # See issue #4932
@@ -224,17 +202,12 @@ class IPyTesProgram(TestProgram):
         TestProgram.__init__(self, *args, **kwargs)
 unittest.main = IPyTesProgram
 
-# Patch ipykernel to avoid errors when setting the Qt5 Matplotlib
-# backemd
-# Fixes Issue 6091
-import ipykernel
-import IPython
-if LooseVersion(ipykernel.__version__) <= LooseVersion('4.7.0'):
-    if ((PY2 and LooseVersion(IPython.__version__) >= LooseVersion('5.5.0')) or
-        (not PY2 and LooseVersion(IPython.__version__) >= LooseVersion('6.2.0'))
-       ):
-        from ipykernel import eventloops
-        eventloops.loop_map['qt'] = eventloops.loop_map['qt5']
+# Ignore some IPython/ipykernel warnings
+try:
+    warnings.filterwarnings(action='ignore', category=DeprecationWarning,
+                            module='ipykernel.ipkernel')
+except:
+    pass
 
 
 #==============================================================================
@@ -278,6 +251,35 @@ except:
     pass
 
 
+# =============================================================================
+# Multiprocessing adjustments
+# =============================================================================
+# This patch is only needed on Windows and Python 3
+if os.name == 'nt' and not PY2:
+    # This could fail with changes in Python itself, so we protect it
+    # with a try/except
+    try:
+        import multiprocessing.spawn
+        _old_preparation_data = multiprocessing.spawn.get_preparation_data
+
+        def _patched_preparation_data(name):
+            """
+            Patched get_preparation_data to work when all variables are
+            removed before execution.
+            """
+            try:
+                return _old_preparation_data(name)
+            except AttributeError:
+                main_module = sys.modules['__main__']
+                # Any string for __spec__ does the job
+                main_module.__spec__ = ''
+                return _old_preparation_data(name)
+
+        multiprocessing.spawn.get_preparation_data = _patched_preparation_data
+    except Exception:
+        pass
+
+
 #==============================================================================
 # Pdb adjustments
 #==============================================================================
@@ -286,7 +288,14 @@ class SpyderPdb(pdb.Pdb):
     send_initial_notification = True
     starting = True
 
-    def set_spyder_breakpoints(self):
+    # --- Methods overriden by us
+    def preloop(self):
+        """Ask Spyder for berkpoints before the first prompt is created."""
+        if self.starting:
+            get_ipython().kernel._ask_spyder_for_breakpoints()
+
+    # --- Methods defined by us
+    def set_spyder_breakpoints(self, breakpoints):
         self.clear_all_breaks()
         #------Really deleting all breakpoints:
         for bp in bdb.Breakpoint.bpbynumber:
@@ -296,22 +305,35 @@ class SpyderPdb(pdb.Pdb):
         bdb.Breakpoint.bplist = {}
         bdb.Breakpoint.bpbynumber = [None]
         #------
-        from spyder.config.main import CONF
-        CONF.load_from_ini()
-        if CONF.get('run', 'breakpoints/enabled', True):
-            breakpoints = CONF.get('run', 'breakpoints', {})
-            i = 0
-            for fname, data in list(breakpoints.items()):
+        i = 0
+        for fname, data in list(breakpoints.items()):
+            if osp.isfile(fname):
                 for linenumber, condition in data:
                     i += 1
                     self.set_break(self.canonic(fname), linenumber,
                                    cond=condition)
 
+        # Jump to first breakpoint.
+        # Fixes issue 2034
+        if self.starting:
+            # Only run this after a Pdb session is created
+            self.starting = False
+
+            # Get all breakpoints for the file we're going to debug
+            frame = self.curframe
+            lineno = frame.f_lineno
+            breaks = self.get_file_breaks(frame.f_code.co_filename)
+
+            # Do 'continue' if the first breakpoint is *not* placed
+            # where the debugger is going to land.
+            # Fixes issue 4681
+            if breaks and lineno != breaks[0] and osp.isfile(fname):
+                get_ipython().kernel.pdb_continue()
+
     def notify_spyder(self, frame):
         if not frame:
             return
 
-        from IPython.core.getipython import get_ipython
         kernel = get_ipython().kernel
 
         # Get filename and line number of the current frame
@@ -323,21 +345,6 @@ class SpyderPdb(pdb.Pdb):
                 pass
         lineno = frame.f_lineno
 
-        # Jump to first breakpoint.
-        # Fixes issue 2034
-        if self.starting:
-            # Only run this after a Pdb session is created
-            self.starting = False
-
-            # Get all breakpoints for the file we're going to debug
-            breaks = self.get_file_breaks(frame.f_code.co_filename)
-
-            # Do 'continue' if the first breakpoint is *not* placed
-            # where the debugger is going to land.
-            # Fixes issue 4681
-            if breaks and lineno != breaks[0] and osp.isfile(fname):
-                kernel.pdb_continue()
-
         # Set step of the current frame (if any)
         step = {}
         if isinstance(fname, basestring) and isinstance(lineno, int):
@@ -408,7 +415,7 @@ def user_return(self, frame, return_value):
 
 @monkeypatch_method(pdb.Pdb, 'Pdb')
 def interaction(self, frame, traceback):
-    if frame is not None and "sitecustomize.py" in frame.f_code.co_filename:
+    if frame is not None and "spydercustomize.py" in frame.f_code.co_filename:
         self.run('exit')
     else:
         self.setup(frame, traceback)
@@ -434,21 +441,19 @@ def _cmdloop(self):
                    "For copying text while debugging, use Ctrl+Shift+C",
                    file=self.stdout)
 
+
 @monkeypatch_method(pdb.Pdb, 'Pdb')
 def reset(self):
     self._old_Pdb_reset()
-
-    from IPython.core.getipython import get_ipython
     kernel = get_ipython().kernel
     kernel._register_pdb_session(self)
-    self.set_spyder_breakpoints()
 
 
 #XXX: notify spyder on any pdb command (is that good or too lazy? i.e. is more
 #     specific behaviour desired?)
 @monkeypatch_method(pdb.Pdb, 'Pdb')
 def postcmd(self, stop, line):
-    if line != "!get_ipython().kernel._set_spyder_breakpoints()":
+    if "_set_spyder_breakpoints" not in line:
         self.notify_spyder(self.curframe)
     return self._old_Pdb_postcmd(stop, line)
 
@@ -499,14 +504,25 @@ class UserModuleReloader(object):
     def __init__(self, namelist=None, pathlist=None):
         if namelist is None:
             namelist = []
-        spy_modules = ['sitecustomize', 'spyder', 'spyderplugins']
+        else:
+            try:
+                namelist = namelist.split(',')
+            except Exception:
+                namelist = []
+
+        # Spyder modules
+        spy_modules = ['spyder_kernels']
+
+        # Matplotlib modules
         mpl_modules = ['matplotlib', 'tkinter', 'Tkinter']
+
         # Add other, necessary modules to the UMR blacklist
         # astropy: see issue 6962
         # pytorch: see issue 7041
         # fastmat: see issue 7190
         # pythoncom: see issue 7190
-        other_modules = ['pytorch', 'pythoncom']
+        # tensorflow: see issue 8697
+        other_modules = ['pytorch', 'pythoncom', 'tensorflow']
         if PY2:
             py2_modules = ['astropy', 'fastmat']
             other_modules = other_modules + py2_modules
@@ -514,44 +530,150 @@ class UserModuleReloader(object):
 
         if pathlist is None:
             pathlist = []
-        self.pathlist = pathlist
+        self.pathlist = self.create_pathlist(pathlist)
+
+        # List of previously loaded modules
         self.previous_modules = list(sys.modules.keys())
 
-    def is_module_blacklisted(self, modname, modpath):
-        if HAS_CYTHON:
-            # Don't return cached inline compiled .PYX files
-            return True
-        for path in [sys.prefix]+self.pathlist:
-            if modpath.startswith(path):
-                return True
-        else:
-            return set(modname.split('.')) & set(self.namelist)
+        # List of module names to reload
+        self.modnames_to_reload = []
 
-    def run(self, verbose=False):
+        # Activate Cython support
+        self.has_cython = False
+        self.activate_cython()
+
+        # Check if the UMR is enabled or not
+        enabled = os.environ.get("SPY_UMR_ENABLED", "")
+        self.enabled = enabled.lower() == "true"
+
+        # Check if the UMR should print the list of reloaded modules or not
+        verbose = os.environ.get("SPY_UMR_VERBOSE", "")
+        self.verbose = verbose.lower() == "true"
+
+    def create_pathlist(self, initial_pathlist):
         """
-        Del user modules to force Python to deeply reload them
+        Add to pathlist Python library paths to be skipped from module
+        reloading.
+        """
+        try:
+            paths = sysconfig.get_paths()
+            lib_paths = [paths['stdlib'],
+                         paths['purelib'],
+                         paths['scripts'],
+                         paths['data']]
+
+            return initial_pathlist + lib_paths
+        except Exception:
+            return initial_pathlist
+
+    def is_module_reloadable(self, module, modname):
+        """Decide if a module is reloadable or not."""
+        if self.has_cython:
+            # Don't return cached inline compiled .PYX files
+            return False
+        else:
+            if (self.is_module_in_pathlist(module) or
+                    self.is_module_in_namelist(modname)):
+                return False
+            else:
+                return True
+
+    def is_module_in_namelist(self, modname):
+        """Decide if a module can be reloaded or not according to its name."""
+        return set(modname.split('.')) & set(self.namelist)
+
+    def is_module_in_pathlist(self, module):
+        """Decide if a module can be reloaded or not according to its path."""
+        modpath = getattr(module, '__file__', None)
+
+        # Skip module according to different criteria
+        if modpath is None:
+            # *module* is a C module that is statically linked into the
+            # interpreter. There is no way to know its path, so we
+            # choose to ignore it.
+            return True
+        elif any([p in modpath for p in self.pathlist]):
+            # We don't want to reload modules that belong to the
+            # standard library or installed to site-packages,
+            # just modules created by the user.
+            return True
+        elif not os.name == 'nt':
+            # Module paths containing the strings below can be ihherited
+            # from the default Linux installation or Homebrew in a
+            # virtualenv.
+            patterns = [r'^/usr/lib.*',
+                        r'^/usr/local/lib.*',
+                        r'^/usr/.*/dist-packages/.*',
+                        r'^/Library/.*'
+            ]
+
+            if [p for p in patterns if re.search(p, modpath)]:
+                return True
+            else:
+                return False
+        else:
+            return False
+
+    def activate_cython(self):
+        """
+        Activate Cython support.
+
+        We need to run this here because if the support is
+        active, we don't to run the UMR at all.
+        """
+        run_cython = os.environ.get("SPY_RUN_CYTHON") == "True"
+
+        if run_cython:
+            try:
+                __import__('Cython')
+                self.has_cython = True
+            except Exception:
+                pass
+
+            if self.has_cython:
+                # Import pyximport to enable Cython files support for
+                # import statement
+                import pyximport
+                pyx_setup_args = {}
+
+                # Add Numpy include dir to pyximport/distutils
+                try:
+                    import numpy
+                    pyx_setup_args['include_dirs'] = numpy.get_include()
+                except Exception:
+                    pass
+
+                # Setup pyximport and enable Cython files reload
+                pyximport.install(setup_args=pyx_setup_args,
+                                  reload_support=True)
+
+    def run(self):
+        """
+        Delete user modules to force Python to deeply reload them
 
         Do not del modules which are considered as system modules, i.e.
         modules installed in subdirectories of Python interpreter's binary
         Do not del C modules
         """
-        log = []
+        self.modnames_to_reload = []
         for modname, module in list(sys.modules.items()):
             if modname not in self.previous_modules:
-                modpath = getattr(module, '__file__', None)
-                if modpath is None:
-                    # *module* is a C module that is statically linked into the
-                    # interpreter. There is no way to know its path, so we
-                    # choose to ignore it.
-                    continue
-                if not self.is_module_blacklisted(modname, modpath):
-                    log.append(modname)
+                # Decide if a module can be reloaded or not
+                if self.is_module_reloadable(module, modname):
+                    self.modnames_to_reload.append(modname)
                     del sys.modules[modname]
-        if verbose and log:
-            _print("\x1b[4;33m%s\x1b[24m%s\x1b[0m"\
-                   % ("Reloaded modules", ": "+", ".join(log)))
+                else:
+                    continue
 
-__umr__ = None
+        # Report reloaded modules
+        if self.verbose and self.modnames_to_reload:
+            modnames = self.modnames_to_reload
+            _print("\x1b[4;33m%s\x1b[24m%s\x1b[0m"\
+                   % ("Reloaded modules", ": "+", ".join(modnames)))
+
+
+__umr__ = UserModuleReloader(namelist=os.environ.get("SPY_UMR_NAMELIST",
+                                                     None))
 
 
 #==============================================================================
@@ -561,7 +683,6 @@ def clear_post_mortem():
     """
     Remove the post mortem excepthook and replace with a standard one.
     """
-    from IPython.core.getipython import get_ipython
     ipython_shell = get_ipython()
     ipython_shell.set_custom_exc((), None)
 
@@ -572,8 +693,6 @@ def post_mortem_excepthook(type, value, tb):
     mortem debugging.
     """
     clear_post_mortem()
-
-    from IPython.core.getipython import get_ipython
     ipython_shell = get_ipython()
     ipython_shell.showtraceback((type, value, tb))
     p = pdb.Pdb(ipython_shell.colors)
@@ -602,7 +721,6 @@ def set_post_mortem():
     """
     Enable the post mortem debugging excepthook.
     """
-    from IPython.core.getipython import get_ipython
     def ipython_post_mortem_debug(shell, etype, evalue, tb,
                tb_offset=None):
         post_mortem_excepthook(etype, evalue, tb)
@@ -620,7 +738,6 @@ if "SPYDER_EXCEPTHOOK" in os.environ:
 #==============================================================================
 def _get_globals():
     """Return current namespace"""
-    from IPython.core.getipython import get_ipython
     ipython_shell = get_ipython()
     return ipython_shell.user_ns
 
@@ -638,16 +755,9 @@ def runfile(filename, args=None, wdir=None, namespace=None, post_mortem=False):
         # UnicodeError, TypeError --> eventually raised in Python 2
         # AttributeError --> systematically raised in Python 3
         pass
-    global __umr__
-    if os.environ.get("SPY_UMR_ENABLED", "").lower() == "true":
-        if __umr__ is None:
-            namelist = os.environ.get("SPY_UMR_NAMELIST", None)
-            if namelist is not None:
-                namelist = namelist.split(',')
-            __umr__ = UserModuleReloader(namelist=namelist)
-        else:
-            verbose = os.environ.get("SPY_UMR_VERBOSE", "").lower() == "true"
-            __umr__.run(verbose=verbose)
+
+    if __umr__.enabled:
+        __umr__.run()
     if args is not None and not isinstance(args, basestring):
         raise TypeError("expected a character buffer object")
     if namespace is None:
@@ -667,10 +777,9 @@ def runfile(filename, args=None, wdir=None, namespace=None, post_mortem=False):
         os.chdir(wdir)
     if post_mortem:
         set_post_mortem()
-    if HAS_CYTHON:
+    if __umr__.has_cython:
         # Cython files
         with io.open(filename, encoding='utf-8') as f:
-            from IPython.core.getipython import get_ipython
             ipython_shell = get_ipython()
             ipython_shell.run_cell_magic('cython', '', f.read())
     else:
